-- Stack object as a coinductive type

data Stack a = Stack 
  { top  :: Maybe a
  , pop  :: Stack a
  , push :: a -> Stack a
  } 

-- circular auxiliary program to construct stacks 
push' :: Stack a -> a -> Stack a
push' s a = s'
  where s' = Stack (Just a) s (push' s')

-- the empty stack
empty :: Stack a
empty = Stack Nothing empty (push' empty)

test  = top (pop (flip push 0 (flip push 1 empty)))  -- Just 1
test' = top (flip push 0 (pop (flip push 1 empty)))  -- Just 0

-- unwinding the circle


-- loop2 :: (forall j<i. Stack j -> a -> Stack j) -> Stack i -> a -> Stack i
loop2 :: (Stack a -> a -> Stack a) -> Stack a -> a -> Stack a
loop2 push2 s a = Stack (Just a) s (push2 (loop2 push2 s a))
-- loop2 ($ i) push2 s a = Stack (Just a) s (push2 i (loop2 i push2 s a))

-- push2 :: Stack i -> a -> Stack i
push2 :: Stack a -> a -> Stack a
push2 = loop2 push2
-- push2 ($ i) = loop2 i (push2 i)

empty2 :: Stack a
empty2 = Stack Nothing empty2 (push2 empty2)
-- empty2 ($ i) = Stack Nothing (empty2 i) (push2 i (empty2 i))

test2  = top (pop (flip push 0 (flip push 1 empty2)))  -- Just 1
test2' = top (flip push 0 (pop (flip push 1 empty2)))  -- Just 0

-- Anton Setzer: push' with lists ------------------------------------

head' :: [a] -> Maybe a
head' []       = Nothing
head' (a : as) = Just a 

push3 :: [a] -> Stack a
push3 l = Stack (head' l) (push3 (tail l)) (\ a -> push3 (a : l))

empty3 :: Stack a
empty3 = push3 []

test3  = top (pop (flip push 0 (flip push 1 empty3)))  -- Just 1
test3' = top (flip push 0 (pop (flip push 1 empty3)))  -- Just 0

